from setuptools import setup, find_packages
from PROJECT import PROJECT


# =====================================================================================================================
# VERSION = (0, 0, 1)   # use find_packages to keep all internal pkgs for pypi
VERSION = (0, 0, 2)   # fix ability to read russian text in readme


# =====================================================================================================================
with open("README.md", mode="r", encoding="utf8") as f:
    readme = f.read()


packages = [PROJECT.NAME_IMPORT, ]
pkgs_internal = find_packages(where=PROJECT.NAME_IMPORT)
for name in pkgs_internal:
    packages.append(f"{PROJECT.NAME_IMPORT}.{name}")


# =====================================================================================================================
setup(
  version=PROJECT.VERSION_STR,
  description=PROJECT.DESCRIPTION_SHORT,
  keywords=PROJECT.KEYWORDS,
  classifiers=[
    # "Topic :: ________________",
    *PROJECT.CLASSIFIERS_TOPICS_ADD,

    # "Framework :: ",
    "Topic :: Software Development :: Libraries :: Python Modules",
    # "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.11",
    "Operating System :: OS Independent",
    # "Environment :: Console",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "Typing :: Typed",
  ],

  name=PROJECT.NAME_IMPORT,
  author=PROJECT.AUTHOR_NAME,
  author_email=PROJECT.AUTHOR_EMAIL,
  long_description=readme,
  long_description_content_type="text/markdown",

  url=PROJECT.AUTHOR_HOMEPAGE,  # HOMEPAGE
  project_urls={
    # "Documentation": f"https://github.com/centroid457/{NAME}/blob/main/GUIDE.md",
    "Source": f"https://github.com/centroid457/{PROJECT.NAME_IMPORT}",
  },

  packages=packages,
  install_requires=[],
  python_requires=">=3.6"
)


# =====================================================================================================================



# """
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# # TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# but maybe leave it on!
# """
#
# from typing import *
#
# # from base_aux.funcs import *
# from base_aux.classes import Valid
#
#
# # =====================================================================================================================
# TYPE__RESULT_CUM_STEP = Union[bool, Any, "Valid"]     # DONT USE CALLABLES!!! Valid is most PREFIRED! specially in case of EXX!!
# TYPE__RESULT_CUM_STEPS = Union[TYPE__RESULT_CUM_STEP, list[TYPE__RESULT_CUM_STEP]]
#
#
# # =====================================================================================================================
# class ResultStep:
#     """
#     # TODO: this is just as idea! but just create clear correct log msgs and this is will be enough!
#     GOAL
#     ----
#     item for steps in sequence
#
#     CREATED SPECIALLY FOR
#     ---------------------
#     for ResultCum as internal aux
#
#     CONSTRAINTS
#     -----------
#
#     BEST USAGE
#     ----------
#     """
#     pass
#
#
# # =====================================================================================================================
# class ResultCum:
#     """
#     GOAL
#     ----
#     apply cumulated result with log lines
#
#     CREATED SPECIALLY FOR
#     ---------------------
#     testplans
#
#     REPLACE: ResultExpected
#
#     CONSTRAINTS
#     -----------
#     1. dont use callables in steps! only final result (as expression) or Valid as most preferred
#     2. if no steps - return True!
#
#     BEST USAGE
#     ----------
#         # some example
#         def meth() -> ResultCum:
#             result_cum = ResultCum()
#
#             result_step = 1+1 == 2
#             result_cum.result__apply_step(result_step)
#             result_cum.log_lines__add('check 1+1 == 2')
#
#             result_step = 1+2 == 4
#             result_cum.result__apply_step(result_step)
#             result_cum.log_lines__add('check 1+2 == 4')
#
#             result_cum.finish()
#             return result_cum
#
#     """
#
#     TITLE: str = ""
#     # COMMENT: str = ""
#
#     result: bool = True
#     finished: bool | None = None    # as help to see if process is finished - maybe need deprecate!
#
#     LOG_LINES: list[str]
#     STEP_HISTORY: list[tuple[TYPE__RESULT_CUM_STEP, bool]]     # as history! result+step
#
#     def __init__(self, title: str = None):        #, steps=None):  #dont use steps here
#         if title:
#             self.TITLE = title
#
#         self.clear()
#
#     def clear(self) -> None:
#         self.result = True
#         self.finished = None
#
#         self.LOG_LINES = []
#         self.STEP_HISTORY = []
#
#     def result__apply_step(self, step: TYPE__RESULT_CUM_STEPS, cumulate: bool | Any = True, msg: Union[None, str, list[str]] = None) -> bool:
#         """
#
#         :param step:
#         :param cumulate: False - if you dont mind step result! but want to keep it in log
#         :param msg:
#         :return:
#         """
#         if isinstance(step, list):
#             # LIST ------------------------------------------
#             for step_i in step:
#                 if not self.result__apply_step(step_i, cumulate):
#                     break
#         else:
#             # SINGLE ------------------------------------------
#             # TODO: if isinstance(step, ResultCum):   - what do i need to do here?? decide!
#             if isinstance(step, Valid):
#                 if not step.finished:
#                     step.run()
#                 self.log_lines__add(str(step))
#
#         if msg:
#             self.log_lines__add(msg)
#
#         step_result = bool(step)
#         self.STEP_HISTORY.append((step_result, step))
#
#         if bool(cumulate):
#             if self.result is None:
#                 self.result = step_result
#             else:
#                 self.result &= step_result
#
#         return self.result
#
#     def log_lines__add(self, new: Union[str, list[str]]) -> None:
#         if isinstance(new, list):
#             self.LOG_LINES.extend(new)
#         if isinstance(new, str):
#             self.LOG_LINES.append(new)
#
#     def finish(self) -> None:
#         self.finished = True
#
#     def __bool__(self) -> bool:
#         return bool(self.result)
#
#     def __str__(self):
#         result = f"{self.__class__.__name__}(result={self.result},finished={self.finished},TITLE={self.TITLE},LOG_LINES:"
#         for i, line in enumerate(self.LOG_LINES):
#             result += f"\n{i:->5d}:{line}"
#         return result
#
#     def __repr__(self):
#         return str(self)
#
#     def step_last__get(self) -> Any | NoReturn:
#         return self.STEP_HISTORY[-1][1]
#
#     def log_last__get(self) -> str | None:
#         try:
#             return self.LOG_LINES[-1]
#         except:
#             pass
#
#
# # =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT


# =====================================================================================================================
# TEMPLATE
# from .main import (
#     # BASE
#     EXACT_OBJECTS,
#
#     # AUX
#
#     # TYPES
#
#     # EXX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .history import HistoryIO
from .parser import LineParsed
from .serial_client import (
    # BASE
    SerialClient,
    # AUX
    # TYPES
    TYPE__ADDRESS,
    TYPE__RW_ANSWER_SINGLE,
    TYPE__RW_ANSWER,
    Type__WrReturn,
    Type__AddressAutoAcceptVariant,
    # EXX
    Exx_SerialAddress_NotApplyed,
    Exx_SerialAddress_NotExists,
    Exx_SerialAddresses_NoVacant,
    Exx_SerialAddresses_NoAutodetected,
    Exx_SerialAddress_AlreadyOpened,
    Exx_SerialAddress_AlreadyOpened_InOtherObject,
    Exx_SerialAddress_OtherError,
    Exx_SerialRead_NotFullLine,
    Exx_SerialRead_FailPattern,
    Exx_SerialRead_FailDecoding,
    Exx_SerialPL2303IncorrectDriver,
)
from .serial_server import (
    # BASE
    SerialServer_Base,
    SerialServer_Echo,
    SerialServer_Example,
    # AUX
    AnswerVariants,
    # TYPES
    TYPE__CMD_RESULT,
    # EXX
)
from .serial_derivatives import (
    # BASE
    SerialClient_FirstFree,
    SerialClient_FirstFree_Shorted,
    SerialClient_FirstFree_Paired,
    SerialClient_FirstFree_AnswerValid,
    SerialClient_Emulated,
    # AUX
    # TYPES
    # EXX
)

from base_aux.requirements import ReqCheckStr_Os
if ReqCheckStr_Os.bool_if__LINUX():
    from .i2c_client import (
        # BASE
        BusI2c,
        # AUX
        Patterns,
        # TYPES
        # EXX
        Exx_I2cConnection,
    )


# =====================================================================================================================



from typing import *
from . import Default


# =====================================================================================================================
"""
HOW to work with ARGS in your funcs
---------------------
usually and maybe often when we passing ARGS it assumed as we will pass tuple/list or direct value
"""


# =====================================================================================================================
class ArgsEmpty(Default):
    """
    THIS IS JUST A PARTIAL CASE FOR ARGS Default

    DEPRECATE ???
    ---------
    USE DIRECTLY THE CONSTANT VALUE ()!!!
    its clear and

    GOAL
    ----
    explicit pass
    resolve not passed parameters in case of None VALUE!

    special object used as VALUE to show that parameter was not passed!
    dont pass it directly! keep it only as default parameter in class and in methods instead of None Value!
    it used only in special cases! not always even in one method!!!

    SAME AS
    -------
    value_explicit.ValueNotExist but just blank collection
    """
    def __init__(self):
        super().__init__(source=())

    # # NOTE: this is not nesessory!!! and even need not to use for sure!
    # @classmethod
    # def __str__(cls):
    #     return "()"
    #
    # @classmethod
    # def __repr__(cls):
    #     return str(cls)
    #
    # @classmethod
    # def __len__(cls):
    #     return 0
    #
    # @classmethod
    # def __bool__(cls):
    #     return False
    #
    # @classmethod
    # def __iter__(cls):
    #     yield from ()


# ---------------------------------------------------------------------------------------------------------------------
TYPE__ARGS_EMPTY = Type[ArgsEmpty] | ArgsEmpty


# =====================================================================================================================
if __name__ == '__main__':
    pass
    # CLS ----------------
    print(ArgsEmpty)    # <class '__main__.ArgsEmpty'>
    print(str(ArgsEmpty))   # <class '__main__.ArgsEmpty'>
    # print(len(ArgsEmpty))   #TypeError: object of type 'type' has no len()
    # print(list(ArgsEmpty))  #TypeError: 'type' object is not iterable
    print()

    # INST ---------------
    victim = ArgsEmpty()
    print(victim)
    print(str(victim))
    print(repr(victim))
    print(bool(victim))
    print(len(victim))
    print(list(victim))


# =====================================================================================================================



from typing import *
import platform

from base_aux.objects import *


# =====================================================================================================================
TYPE__VALUES = Union[str, list[str], dict[str, bool | None]]

TYPE__RESULT_BOOL = Callable[..., bool]
TYPE__RESULT_RAISE = Callable[..., Optional[NoReturn]]


# =====================================================================================================================
class Exx_RequirementCantGetActualValue(Exception):
    """
    """


class Exx_Requirement(Exception):
    """Some of the requirements are not match
    """


# =====================================================================================================================
class _GetattrClassmethod_Meta(type):
    """ability to apply classmethod for __getattr__

    WHY USE IT
    ==========
    cause of direct __getattr__ usage - is not applicable!

        class Cls:
        @classmethod
        def __getattr__(cls, item):
            print(item)

        Cls.hello()

        # RESULT
            Cls.hello()
            ^^^^^^^^^
        AttributeError: type object 'Cls' has no attribute 'hello'

    WHY WE NEED CLASSMETH instead of simple SELFMETHOD
    --------------------------------------------------
    1. ability to use methods without creating instances - its a quick/simple
        from requirements import ReqCheckStr_Os
        ReqCheckStr_Os.raise_if_not__LINUX
    """

    # dont change markers! use exists!
    _MARKER__BOOL_IF: str = "bool_if__"
    _MARKER__BOOL_IF_NOT: str = "bool_if_not__"
    _MARKER__RAISE_IF: str = "raise_if__"
    _MARKER__RAISE_IF_NOT: str = "raise_if_not__"

    def __getattr__(cls, item: str):
        """if no exists attr/meth
        """
        if item.lower().startswith(cls._MARKER__BOOL_IF.lower()):
            attr_name = item.lower().replace(cls._MARKER__BOOL_IF.lower(), "")
            return lambda: cls.check(values=attr_name, _raise=False, _reverse=False, _meet_true=False)
        elif item.lower().startswith(cls._MARKER__BOOL_IF_NOT.lower()):
            attr_name = item.lower().replace(cls._MARKER__BOOL_IF_NOT.lower(), "")
            return lambda: cls.check(values=attr_name, _raise=False, _reverse=True, _meet_true=False)

        elif item.lower().startswith(cls._MARKER__RAISE_IF.lower()):
            attr_name = item.lower().replace(cls._MARKER__RAISE_IF.lower(), "")
            return lambda: not cls.check(values=attr_name, _raise=True, _reverse=True, _meet_true=False) or None
        elif item.lower().startswith(cls._MARKER__RAISE_IF_NOT.lower()):
            attr_name = item.lower().replace(cls._MARKER__RAISE_IF_NOT.lower(), "")
            return lambda: not cls.check(values=attr_name, _raise=True, _reverse=False, _meet_true=False) or None

        else:
            msg = f"[ERROR] META:'{cls.__name__}' CLASS has no attribute '{item}'"
            raise AttributeError(msg)


# =====================================================================================================================
class ReqCheckStr_Base(metaclass=_GetattrClassmethod_Meta):
    """Base class for check exact requirement by string value

    NOTE
    ----
    USUALLY YOU DONT NEED USING IT LIKE INSTANCE!
    just create appropriate class with _GETTER +add bare annotations with markers (see examples like ReqCheckStr_Os)

    VARIANTS for check
    ------------------
    add attributes with bool values (case-insensitive)
        - True - if value is definitely acceptable - Always need at least one True!!!
        - False - if not definitely acceptable
        - None - if requirement is undefined

    SETTINGS
    --------
    :ivar _RAISE: raise in case of inacceptance
    :ivar _MEET_TRUE: you can use requirement class for check only false variant
    :ivar _CHECK_FULLMATCH:
        True - if need fullmatch (but always case-insensitive)
        False - if partial match by finding mentioned values in _value_actual

    :ivar _GETTER: function which will get the exact value to check
    :ivar _sample_actual:
    """
    # SETTINGS -------------------------------------
    _GETTER: Union[Callable[..., Union[str, Any]], Any] = None

    # AUX ---------------------------------------
    _RAISE: bool = True
    _MEET_TRUE: bool = True
    _CHECK_FULLMATCH: bool = True

    # temporary ------------------------------------------
    _value_actual: Optional[str]

    # TODO: add values as dict??? - it would be direct great!

    def __init__(
            self,
            _getter: Callable[..., str] = None,
            _raise: Optional[bool] = None,
            _meet_true: Optional[bool] = None,
            _check_fullmatch: Optional[bool] = None
    ):
        # INIT SETTINGS ----------------------------------
        if _getter is not None:
            self.__class__._GETTER = _getter
        if _raise is not None:
            self._RAISE = _raise
        if _meet_true is not None:
            self._MEET_TRUE = _meet_true
        if _check_fullmatch is not None:
            self._CHECK_FULLMATCH = _check_fullmatch

        # WORK -------------------------------------------
        self.check()

    def __getattr__(self, item):
        """
        apply access to not exists methods from instance! in metaclass we have only access as classmethods!
        """
        # return super().__getattr__(item)    # AttributeError: 'super' object has no attribute '__getattr__'. Did you mean: '__setattr__'?
        return _GetattrClassmethod_Meta.__getattr__(self.__class__, item)

    @classmethod
    def values_acceptance__get(cls) -> dict[str, bool | None]:
        """get settings from class"""
        values = {}
        for attr in dir(cls):
            attr_value = getattr(cls, attr)
            if not attr.startswith("_") and not callable(attr_value) and isinstance(attr_value, (bool, type(None))):
                values.update({attr: attr_value})
        return values

    @classmethod
    def _value_actual__get(cls) -> Union[str, NoReturn]:
        if not cls._GETTER:
            msg = f"[ERROR] incomplete settings [{cls._GETTER=}]"
            raise Exx_RequirementCantGetActualValue(msg)

        if TypeChecker.check__callable_func_meth_inst(cls._GETTER):
            try:
                cls._value_actual = str(cls._GETTER()).lower()
            except Exception as exx:
                raise Exx_RequirementCantGetActualValue(repr(exx))
        else:
            cls._value_actual = str(cls._GETTER).lower()

        return cls._value_actual

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def check(
            cls,
            values: TYPE__VALUES | None = None,
            _raise: Optional[bool] = None,
            _reverse: Optional[bool] = None,    # special for bool_if_not__* like methods
            _meet_true: bool | None = None,
    ) -> Union[bool, NoReturn, None]:
        # SETTINGS -------------------------------------------------------
        if _raise is None:
            _raise = cls._RAISE
        _reverse = _reverse or False

        if _meet_true is None:
            _meet_true = cls._MEET_TRUE

        # VALUES ---------------------------------------------------------
        # use values-1=from class settings -----------
        if values is None:
            values = cls.values_acceptance__get()

        # use values-2=as exact one -----------
        if isinstance(values, str):
            values = {values: True}

        # use values-3=as exact several -----------
        if isinstance(values, list):
            values = dict.fromkeys(values, True)

        # REVERSE ---------------------------------------------------
        if _reverse:
            for value, acceptance in values.items():
                if acceptance in (True, False):
                    values[value] = not acceptance

        # VALUE ACTUAL ---------------------------------------------------
        _value_actual = cls._value_actual__get()

        # WORK -----------------------------------------------------------
        match = None
        _acceptance = None
        for value, _acceptance in values.items():
            match = (
                (cls._CHECK_FULLMATCH and value.lower() == _value_actual.lower())
                or
                (not cls._CHECK_FULLMATCH and value.lower() in _value_actual.lower())
            )
            if match:
                break

        if match:
            acceptance = _acceptance
        else:
            acceptance = not _reverse

        # acceptance --------------
        result = None
        if acceptance is True:
            result = match
        elif acceptance is False:
            result = not match
        elif acceptance is None:
            result = None

        # FINAL --------------
        if _meet_true is True and result is None:
            msg = f"[WARN] value is not MeetTrue [{cls.__name__}/{cls._value_actual=}/req={values}]"
            print(msg)
            if _raise:
                raise Exx_Requirement(msg)
            else:
                return False

        if result in (True, None):
            return result
        else:
            msg = f"[WARN] value is not [{cls.__name__}/{cls._value_actual=}/req={values}]"
            print(msg)
            if _raise:
                raise Exx_Requirement(msg)
            else:
                return False


# =====================================================================================================================
class ReqCheckStr_Os(ReqCheckStr_Base):
    _GETTER: Callable = platform.system
    _MEET_TRUE: bool = False        # need to use class as checker

    LINUX: bool
    WINDOWS: bool

    # DERIVATIVES --------
    bool_if__LINUX: TYPE__RESULT_BOOL
    bool_if__WINDOWS: TYPE__RESULT_BOOL
    bool_if_not__LINUX: TYPE__RESULT_BOOL
    bool_if_not__WINDOWS: TYPE__RESULT_BOOL

    raise_if__LINUX: TYPE__RESULT_RAISE
    raise_if__WINDOWS: TYPE__RESULT_RAISE
    raise_if_not__LINUX: TYPE__RESULT_RAISE
    raise_if_not__WINDOWS: TYPE__RESULT_RAISE


# =====================================================================================================================
class ReqCheckStr_Arch(ReqCheckStr_Base):
    _GETTER: Callable = platform.machine
    _MEET_TRUE: bool = False

    AMD64: bool      # standard PC
    x86_64: bool     # wsl standard
    AARCH64: bool    # raspberry=ARM!

    # DERIVATIVES --------
    raise_if_not__AARCH64: TYPE__RESULT_RAISE


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from base_aux.testplans import *
from base_aux.pyqt import *
from base_aux.classes import *

from .tc import TestCaseBase
from .tm import TpTableModel


# =====================================================================================================================
class TpHlStyles(HlStyles):
    RESULT: HlStyle = HlStyle(
        FORMAT=format_make("", "", "lightGrey"),
        P_ITEMS=[
            "validate_last_bool", "result",
        ],
        P_TEMPLATES=[
            r'.*%s.*',
        ],
    )
    RESULT_TRUE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "lightGreen"),
        P_ITEMS=[
            "True",
        ],
        P_TEMPLATES=[
            r'.*validate_last_bool=%s.*',
            r'.*result[^=]*=%s.*',
            # r'.*valid[^=]*=%s.*',
        ],
    )
    RESULT_FALSE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "pink"),
        P_ITEMS=[
            "False",
        ],
        P_TEMPLATES=[
            r'.*validate_last_bool=%s.*',
            r'.*result[^=]*=%s.*',
            # r'.*valid[^=]*=%s.*',
        ],
    )
    RESULT_FALSE_NOCUM: HlStyle = HlStyle(
        FORMAT=format_make("", "", "yellow"),
        P_ITEMS=[
            "ValidNoCum",
        ],
        P_TEMPLATES=[
            r'.*%s.*validate_last_bool=False.*',
        ],
    )


PTE_RESULTS_EXAMPLE = """
ПРОВЕРКА выделения текста результатов
Valid=None    #без явного bool значения
Result=True   #явное True
Valid=True    #явное True
Valid=False   #явное False
ValidNoCum(123x=False, validate_last_bool=True)
ValidNoCum(123x=True, validate_last_bool=False)  #неважный/неучтенный False

result__startup=True
------------------------------------------------------------
result=Valid(NAME=GET,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<function SerialClient.__getattr__.<locals>.<lambda> at 0x000001E5AABF6FC0>,ARGS__VALUE=('PRSNT',),value_last=0,
...VALIDATE_LINK=0,validate_last=True,
,finished=True,timestamp_last=1730375894.4305248,
------------------------------------------------------------
result__teardown=ValidChains(NAME=,skip_last=False,validate_last_bool=True,
...VALUE_LINK=None,value_last=None,
...VALIDATE_LINK=True,validate_last=True,
,finished=True,timestamp_last=1730375895.0524113,
result__startup=ValidSleep(NAME=Sleep,skip_last=False,validate_last_bool=True,
___0:(START) len=3/timestamp=1730375895.0524113
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=False,
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=True,
___1:ValidNoCum(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=None,
___1:Valid(NAME=DUT.connect__only_if_address_resolved,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<bound method SerialClient.connect__only_if_address_resolved of <DEVICES.ptb.Device object at 0x000001E5AAB4B050>>,value_last=True,
...VALIDATE_LINK=True,validate_last=True,
,finished=True,timestamp_last=1730375895.0524113,
___2:ValidRetry1(NAME=RST,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<function SerialClient.__getattr__.<locals>.<lambda> at 0x000001E5AABF7240>,value_last=OK,
...VALIDATE_LINK=OK,validate_last=True,
,finished=True,timestamp_last=1730375896.3252466,
___3:ValidSleep(NAME=Sleep,skip_last=False,validate_last_bool=True,
...VALUE_LINK=<built-in function sleep>,ARGS__VALUE=(1,),value_last=None,
...VALIDATE_LINK=None,validate_last=True,
,finished=True,timestamp_last=1730375896.9521606,
___4:(FINISH) [result=False]/len=3
------------------------------------------------------------
DETAILS=====================
"""


# =====================================================================================================================
class TpGuiBase(Gui):
    # OVERWRITTEN -----------------------------------
    START = False

    TITLE = "[TestPlan] Title"
    SIZE = (600, 300)

    HL_STYLES = TpHlStyles()

    # NEW -------------------------------------------
    DATA: "TpMultyDutBase"

    def __init__(self, data):
        self.TITLE = f"[TestPlan]{data.STAND_NAME}/{data.STAND_DESCRIPTION[:20]}"
        super().__init__(data)

    # WINDOW ==========================================================================================================
    def wgt_create(self):
        self.TV_create()
        self.PTE_create()
        self.BTN_create()
        self.CB_create()
        self.HL_create()

        # DETAILS -----------------------------------------------------------------------------------------------------

        # layout_details ----------------------------------------------------------------------------------------------
        layout_details = QVBoxLayout()
        layout_details.addWidget(self.BTN_devs_detect)
        layout_details.addWidget(self.BTN_start)
        layout_details.addWidget(self.CB_tp_run_infinit)
        layout_details.addWidget(self.CB_tc_run_single)
        layout_details.addWidget(self.BTN_settings)
        layout_details.addWidget(self.BTN_clear_all)
        layout_details.addWidget(self.PTE)

        # layout_main -------------------------------------------------------------------------------------------------
        layout_main = QHBoxLayout()
        layout_main.addWidget(self.TV)
        layout_main.addLayout(layout_details)
        self.setLayout(layout_main)

    # WGTS ============================================================================================================
    def BTN_create(self) -> None:
        self.BTN_devs_detect = QPushButton("определить устройства")
        self.BTN_devs_detect.setCheckable(False)

        self.BTN_start = QPushButton("ЗАПУСК ТЕСТОВ")
        self.BTN_start.setCheckable(True)

        self.BTN_settings = QPushButton("настройки")
        self.BTN_settings.setCheckable(True)

        self.BTN_clear_all = QPushButton("очистить результаты")

    def CB_create(self) -> None:
        self.CB_tp_run_infinit = QCheckBox("бесконечный цикл")
        self.CB_tc_run_single = QCheckBox("запустить только выбранный тесткейс")

        # SETTINGS -------------------------
        # self.CB_tp_run_infinit.setText("CB_text")
        # self.CB_tp_run_infinit.setText("CB_text")

    def PTE_create(self) -> None:
        self.PTE = QPlainTextEdit()

        # METHODS ORIGINAL ---------------------------------
        # self.PTE.setEnabled(True)
        # self.PTE.setUndoRedoEnabled(True)
        # self.PTE.setReadOnly(True)
        # self.PTE.setMaximumBlockCount(15)

        # self.PTE.clear()
        self.PTE.appendPlainText(PTE_RESULTS_EXAMPLE)

        # self.PTE.appendHtml("")
        # self.PTE.anchorAt(#)
        # self.PTE.setSizeAdjustPolicy(#)

        # METHODS COMMON -----------------------------------
        self.PTE.setFont(QFont("Calibri (Body)", 7))

    def TV_create(self):
        # TODO: move examples to pyqtTemplate!
        self.TM = TpTableModel(self.DATA)

        self.TV = QTableView()
        self.TV.setModel(self.TM)
        self.TV.setSelectionMode(QTableView.SingleSelection)

        # self.TV.setStyleSheet("gridline-color: rgb(255, 0, 0)")
        # self.TV.setMinimumSize(400, 300)
        # self.TV.setShowGrid(True)
        # self.TV.setFont(QFont("Calibri (Body)", 7))
        # self.TV.setSortingEnabled(True)     # enable sorting
        self.TV.resizeColumnsToContents()   # set column width to fit contents
        # self.TV.setColumnWidth(0, 100)

        #
        hh = self.TV.horizontalHeader()
        # hh.setSectionHidden(self.TM.ADDITIONAL_COLUMNS - 1, True)
        hh.setSectionsClickable(False)
        # hh.setStretchLastSection(True)

        # self.TV.selectRow(1)      # not working
        # self.TV.selectColumn(2)   # not working

        # self.TV.setSelectionModel(QItemSelection().select())

    # SLOTS ===========================================================================================================
    def slots_connect(self):
        self.CB_tp_run_infinit.stateChanged.connect(self.CB_tp_run_infinit__changed)
        self.CB_tc_run_single.stateChanged.connect(self.CB_tc_run_single__changed)

        self.BTN_start.toggled.connect(self.BTN_start__toggled)
        self.BTN_settings.toggled.connect(self.BTN_settings__toggled)
        self.BTN_devs_detect.clicked.connect(self.BTN_devs_detect__clicked)
        self.BTN_clear_all.clicked.connect(self.BTN_clear_all__clicked)

        self.DATA.signal__tp_finished.connect(lambda: self.BTN_start.setChecked(False))
        self.DATA.signal__tp_finished.connect(self.TM._data_reread)

        TestCaseBase.signals.signal__tc_state_changed.connect(lambda _: self.TM._data_reread())

        self.TV.selectionModel().selectionChanged.connect(self.TV_selectionChanged)
        self.TV.horizontalHeader().sectionClicked.connect(self.TV_hh_sectionClicked)

    # -----------------------------------------------------------------------------------------------------------------
    def CB_tp_run_infinit__changed(self, state: Optional[int] = None) -> None:
        """
        :param state:
            0 - unchecked
            1 - halfCHecked (only if isTristate)
            2 - checked (even if not isTristate)
        """
        self.DATA.TP_RUN_INFINIT = bool(state)

    def CB_tc_run_single__changed(self, state: Optional[int] = None) -> None:
        """
        :param state:
            0 - unchecked
            1 - halfCHecked (only if isTristate)
            2 - checked (even if not isTristate)
        """
        if state:
            self.DATA._TC_RUN_SINGLE = True
        else:
            self.DATA._TC_RUN_SINGLE = False
            if not self.DATA.isRunning():
                self.DATA.tc_active = None

        self.TM._data_reread()

    def BTN_start__toggled(self, state: Optional[bool] = None) -> None:
        # print(f"btn {state=}")
        if self.DATA.isRunning():
            state = False

        if state:
            self.DATA.start()
            self.TM._data_reread()
        elif state is False:
            # if not self.DATA.isFinished():
            self.DATA.terminate()
            self.TM._data_reread()

    def BTN_settings__toggled(self, state: Optional[bool] = None) -> None:
        # print(f"BTN_select_tc_on_duts__toggled {state=}")
        # self.TV.horizontalHeader().setSectionHidden(self.TM.ADDITIONAL_COLUMNS - 1, not state)
        self.TV.horizontalHeader().setSectionsClickable(state)
        self.TM.open__settings = state
        self.TM._data_reread()

    def BTN_devs_detect__clicked(self) -> None:
        self.DATA.DEVICES__BREEDER_CLS.group_call__("address_forget")
        self.DATA.DEVICES__BREEDER_CLS.CLS_LIST__DUT.ADDRESSES__SYSTEM.clear()
        self.TM._data_reread()
        self.DATA.DEVICES__BREEDER_CLS.group_call__("address__resolve")    # MOVE TO THREAD??? no! not so need!
        self.TM._data_reread()

    def BTN_clear_all__clicked(self) -> None:
        self.DATA.tcs_clear()
        self.TM._data_reread()

    def TV_hh_sectionClicked(self, index: int) -> None:
        if index == self.TM.HEADERS.STARTUP_CLS:
            pass

        if index == self.TM.HEADERS.TEARDOWN_CLS:
            pass

        if index in self.TM.HEADERS.DUTS:
            dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[self.TM.HEADERS.DUTS.get_listed_index__by_outer(index)]
            dut.SKIP_reverse()
            self.TM._data_reread()

    def TV_selectionChanged(self, first: QItemSelection, last: QItemSelection) -> None:
        # print("selectionChanged")
        # print(f"{first=}")  # first=<PyQt5.QtCore.QItemSelection object at 0x000001C79A107460>
        # ObjectInfo(first.indexes()[0]).print(_log_iter=True, skip_fullnames=["takeFirst", "takeLast"])

        if not first:
            # when item with noFlag IsSelectable
            return

        index: QModelIndex = first.indexes()[0]

        row = index.row()
        col = index.column()

        tc_cls = list(self.DATA.TCS__CLS)[row]

        if self.DATA._TC_RUN_SINGLE:
            self.DATA.tc_active = tc_cls

        if col == self.TM.HEADERS.STARTUP_CLS:
            self.PTE.setPlainText(str(tc_cls.result__startup_cls))

        if col in self.TM.HEADERS.DUTS:
            dut = self.DATA.DEVICES__BREEDER_CLS.LIST__DUT[col - self.TM.HEADERS.DUTS.START_OUTER]
            self.PTE.setPlainText(tc_cls.TCS__LIST[dut.INDEX].get__results_pretty())

        if col == self.TM.HEADERS.TEARDOWN_CLS:
            self.PTE.setPlainText(str(tc_cls.result__teardown_cls))

        self.TM._data_reread()


# =====================================================================================================================



from .indicators import (
    IndicatorName,
    IndicatorParamsBase,

    IndicatorParams_WMA,
    IndicatorParams_STOCH,
    IndicatorParams_ADX,
    IndicatorParams_MACD,
    IndicatorParams_RSI,
)
from .mt import (
    PrivateMT5,
    MT5,

    Type__Symbol,
    Type__SymbolOpt,
    Type__Tf,
    Type__TfOpt,
    Type__PdSeries,
    Type__IndicatorValues,

    Exx__Mt5Auth,
    Exx__Mt5SymbolName,

)
from .time_series import (
    HistoryShifted_Shrink,
    HistoryShifted_Simple,

    Exx_TimeSeries,
)
from .symbols import (
    Symbols,
)
from .strategy import (
    MonitorBase,
    AlertTradeADX,
    ThreadManagerADX,
    MonitorADX,

    Alert_MapDrawer,
    ThreadManager_MapDrawer_Tf,
    ThreadManager_MapDrawer_Shift,
    IndicatorMapDrawer_Simple,
)



# GOAL
# KEEP STATIC OBJECTS like TYPE__* and Exx__* in separated file.
# make clear importing by resolving circular imports!
from typing import Type, Any, Callable, NoReturn


# =====================================================================================================================
class Exx__AnnotNotDefined(Exception):
    """Exception in case of not defined attribute in instance
    """


class Exx__NumberArithm_NoName(Exception):
    pass


class Exx__GetattrPrefix(Exception):
    pass


class Exx__GetattrPrefix_RaiseIf(Exx__GetattrPrefix):
    pass


class Exx__ValueNotInVariants(Exception):
    pass


class Exx__VariantsIncompatible(Exception):     # TODO: seems need to deprecate it! it is not important!
    pass


class Exx__ValueNotParsed(Exception):
    pass


class Exx__ValueUnitsIncompatible(Exception):
    pass


class Exx__IndexOverlayed(Exception):
    pass


class Exx__IndexNotSet(Exception):
    pass


class Exx__ItemNotExists(Exception):
    """
    not exists INDEX (out of range) or NAME not in defined values
    """
    pass


class Exx__StartOuterNONE_UsedInStackByRecreation(Exception):
    """
    in stack it will be recreate automatically! so dont use in pure single BreederStrSeries!
    """
    pass


class Exx__BreederObjectList_GroupsNotGenerated(Exception):
    pass


class Exx__BreederObjectList_GroupNotExists(Exception):
    pass


class Exx__BreederObjectList_ObjCantAccessIndex(Exception):
    pass


# =====================================================================================================================
TYPE__LAMBDA_CONSTRUCTOR = Type[Any] | Callable[..., Any | NoReturn]
TYPE__LAMBDA_ARGS = tuple[Any, ...]
TYPE__LAMBDA_KWARGS = dict[str, Any]


# =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT
# VERSION = (0, 0, 2)   # del blank lines
# VERSION = (0, 0, 3)   # separate all types/exx into static.py!


# =====================================================================================================================
# TEMPLATE
# from .static import (
#     # TYPES
#     # EXX
# )
# from .main import (
#     # BASE
#     # AUX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .static import (
    # TYPES
    TYPE__PV_DICT,
    TYPE__PATH,
    TYPE__VALUE,
    # EXX
    Exx__FileNotExists,
    Exx__SameKeys,
)
from .attr_loader__0_base import (
    # BASE
    PrivateBase,
    # AUX
)
from .derivatives import (
    # BASE
    PrivateAuth,
    PrivateTgBotAddress,
    # AUX
)
from .attr_loader__1_env import (
    # BASE
    PrivateEnv,
    # AUX
)
from .attr_loader__2_csv import (
    # BASE
    PrivateCsv,
    PrivateAuthCsv,
    PrivateTgBotAddressCsv,
    # AUX
)
from .attr_loader__3_ini import (
    # BASE
    PrivateIni,
    PrivateAuthIni,
    PrivateTgBotAddressIni,
    # AUX
)
from .attr_loader__4_json import (
    # BASE
    PrivateJson,
    PrivateAuthJson,
    PrivateTgBotAddressJson,
    # AUX
)
from .attr_loader__5_auto import (
    # BASE
    PrivateAuto,
    PrivateAuthAuto,
    PrivateTgBotAddressAuto,
    # AUX
)

# =====================================================================================================================
